#ifndef AMR
#define AMR

#include <cmath>
#include <fstream>
#include "fileIOpara.hh"
#include <dune/grid/utility/persistentcontainer.hh>
#include <dune/grid/geometrygrid/grid.hh>

struct RestrictedValue
{
  double value;
  int count;
  RestrictedValue()
  {
    value = 0;
    count = 0;
  }
};

template <class G, class M, class V, class W, class Q>
bool adapting(G& grid, M& Mapper, const V& IndiField, W& data, Q& data2, Dune::MPIHelper& helper, const char IndiType, int lmin, int lmax, const int plotCounter, const double rtol=0.05, const double ctol=0.01) // lmin, lmax < 0 means no limitations
{
  typedef typename G::ctype ct;
  const int dim = G::dimension;
  
  //Setting view types and iterators
  typedef typename G::LeafGridView LeafGridView;
  typedef typename G::LevelGridView LevelGridView;
  typedef typename LeafGridView::template Codim<0>::Iterator LeafElementIterator;
  typedef typename LevelGridView::template Codim<0>::Iterator LevelElementIterator;
    
  //Entities and Intersections types
  typedef typename G::template Codim<0>::Entity ElementEntity;
  typedef typename ElementEntity::Geometry ElementGeometry;
  typedef typename ElementGeometry::GlobalCoordinate gc;
  typedef typename G::template Codim<0>::EntityPointer ElementEntityPointer;

  typedef typename LeafGridView::IntersectionIterator LeafIntersectionIterator;
  typedef typename LeafIntersectionIterator::Intersection Intersection;
  typedef typename Intersection::GlobalCoordinate igc;
  typedef typename Intersection::Geometry IntersectionGeometry;

  //Container with persistent numbering for reorganization of the user data (vector is default)
  typedef Dune::PersistentContainer<G,RestrictedValue> RestrictionMap;

  LeafGridView leafview = grid.leafView();
  //Iterators
  LeafElementIterator ebegin = leafview.template begin<0>();
  LeafElementIterator eend = leafview.template end<0>();


  V indicator;
  int marked = 0;

  //print intersection info on parallel files
  std::ofstream bndFileStream;
  openFileIOPara(helper, "./output/bndIntersections_", ".dat", bndFileStream);
  std::ofstream intFileStream;
  openFileIOPara(helper, "./output/Intersections_", ".dat", intFileStream);
  intFileStream.setf(std::ios_base::left);

  //---------------------//
  //INDICATOR COMPUTATION//
  //---------------------//
  //Choose marking logic: G = use IndiField Gradient as Indicator, F = use IndiField as indicator directly
  switch (IndiType) {
  case 'F':
    indicator = IndiField;
    break;
  case 'G':
    {
      indicator.resize(IndiField.size(),0);
      for(LeafElementIterator it = ebegin; it != eend; ++it)
	{
	  Dune::FieldVector<ct,dim> gradient(0);
	  double alpha = 0;
	  double buffer = 0;
	  Dune::FieldVector<ct,dim> Buff(0);
	  const int indexi = Mapper.map(*it); //from this cell value can be got
	  const ElementGeometry &geo = it->geometry();
	  const double elementVolume = geo.volume(); //element volume
	  const gc elementCenter = geo.center(); //element position

	  intFileStream << it->partitionType() <<  "Element = " << std::setw(6) << indexi << elementCenter << " "  << std::setw(10) << elementVolume << std::endl;

	  LeafIntersectionIterator isend = leafview.iend(*it);
	  for(LeafIntersectionIterator is = leafview.ibegin(*it); is !=isend; ++is)
	    {
	      const Intersection &intersection = *is;
	      const int indexInsideInt = intersection.indexInInside();
	      const IntersectionGeometry &igeo = intersection.geometry();
	      const double intersectionVolume = igeo.volume(); //intersection area
	      const igc intersectionCenter = igeo.center(); //intersection position
	      const igc normal = is->centerUnitOuterNormal(); //intersection unit normal; methods are available for more generic normal (parametric element!?!!!?!)
	      if( !intersection.neighbor())
		{
		  bndFileStream << "I'm " << helper.rank() << " and this boundary intersection is centered in " << intersectionCenter << std::endl;
		  continue;
		}
	      	      
	      const ElementEntityPointer ptrOutside = intersection.outside();
	      const ElementEntity &outside = *ptrOutside;
	      const gc outsideElementCenter = ptrOutside->geometry().center(); //neighbor position
	      int indexj = Mapper.map(outside); // from this neighbor cell value can be got

	      intFileStream << "Intersection = " << std::setw(6) << indexInsideInt << intersectionCenter << " " << std::setw(10) << intersectionVolume << " " << normal << " " << std::setw(6) << indexj  << " " << outsideElementCenter << std::endl;


	      //WE CAN ENHANCE PERFORMANCE COMPUTING HERE THE CONTRIBUTION OF THE INTERSECTION TO THE GRADIENT IN OUTSIDE
	      alpha = modulo(outsideElementCenter-intersectionCenter)/modulo(outsideElementCenter-elementCenter);
	      buffer = 1.0/elementVolume * ( alpha * IndiField[indexi] + (1.0 - alpha) * IndiField[indexj] ) * intersectionVolume;
	      Buff = normal;
	      Buff *= buffer;
	      gradient += Buff;

	    }
	  indicator[indexi] = modulo(gradient);
	  intFileStream << std::setw(80) << "----------" << std::endl;
	  
	  // //----------------//
	  // //GRADIENT MARKING//
	  // //----------------//
	  // //if refinement is not based on min and/or max values of the indicator field, the marking can happen here
	  // //otherwise a new loop over the elements is needed.
	  // const ElementEntity &entity = *it;
	  // if(indicator[indexi] > rtol)
	  //   grid.mark(1,entity);
	  // else if(indicator[indexi] < ctol)
	  //   grid.mark(-1,entity);
	}

    }
    break;
  default:
    indicator = IndiField;
  }
  //end switch


  closeFileIOPara(bndFileStream);
  closeFileIOPara(intFileStream);

  //plot gradient
  // std::string path = "output";
  // std::string extPath = "";
  // Dune::VTKWriter<LeafGridView> vtkwriter(leafview);
  // vtkwriter.addCellData(indicator,"data");
  // vtkwriter.pwrite("gradientData",path,extPath,Dune::VTKOptions::ascii);
  //end indicator computation

  
  
  //if the indicator requires a loop over all the elements to be defined, a new loop is needed to mark elements
  //-------//
  //MARKING//
  //-------//
  std::ofstream coaFileStream;
  openFileIOPara(helper, "./outputNEW/coarsen_", ".dat", coaFileStream);
  for(LeafElementIterator it = ebegin; it != eend; ++it)
    {
      const int indexi = Mapper.map(*it);
      const ElementEntity &entity = *it;
      if(indicator[indexi] >= rtol && it->level() <= lmax)
	{
	  if(grid.getMark(*it) == 0 )
	    {
	      grid.mark(1,entity);
	      ++marked;
	    }
	  for(LeafIntersectionIterator is = leafview.ibegin(*it); is != leafview.iend(*it); ++is)
	    {
	      const Intersection &intersection = *is;
	      if(intersection.neighbor()){
	  	const ElementEntityPointer ptrOutside = intersection.outside();
	  	const ElementEntity &outside = *ptrOutside;
	  	if(outside.level() < it->level() && grid.getMark(outside) == 0 && grid.getMark(*it) != 0)
	  	  {
	  	    grid.mark(1,outside);
	  	    ++marked;
	  	  }
	      }
	    }
	}
      else if(indicator[indexi] < ctol && it->level() > lmin)
	{
	  for(LeafIntersectionIterator is = leafview.ibegin(*it); is != leafview.iend(*it); ++is)
	    {
	      const Intersection &intersection = *is;
	      if(intersection.neighbor()){
	  	const ElementEntityPointer ptrOutside = intersection.outside();
	  	const ElementEntity &outside = *ptrOutside;
	  	const int indexj = Mapper.map(outside);
	  	if((outside.level() == it->level() && indicator[indexj] > rtol) || (outside.level() > it->level() && indicator[indexj] > ctol))
	  	  continue;
	  	else if( grid.getMark(*it) == 0)
	  	  {
		    bool test;
		    test = grid.mark(-1,entity);
		    coaFileStream << indexi << " " << test << std::endl;
		    ++marked;
		  }
	      }
	    }
	}
    }
  closeFileIOPara(coaFileStream);

  if(marked==0)
    return false; 

  grid.preAdapt();

  //-------------------------------//
  //MAPPING TO PERSISTENT CONTAINER//
  //-------------------------------//
  //put user data in a container using an IndexSet
  RestrictionMap restrictionmap(grid,0);
  RestrictionMap restrictionmap2(grid,0);
  for(int level = grid.maxLevel(); level>=0; --level)
    {
      LevelGridView levelView = grid.levelView(level);
      for(LevelElementIterator it = levelView.template begin<0>(); it != levelView.template end<0>(); ++it)
	{
	  RestrictedValue& rv = restrictionmap[*it];
	  //put leaf values in the persistent container
	  if(it->isLeaf())
	    {
	      int indexi = Mapper.map(*it);
	      rv.value = data[indexi];
	      rv.count = 1;
	    }
	  
	  //average in father as example to test paralell amr
	  //ATTENTION: for parametrized elements the weight of a son should depend on is relative to the father volume
	  if(it->level() > 0)
	    {
	      ElementEntityPointer ep = it->father();
	      RestrictedValue& rvf = restrictionmap[*ep];
	      rvf.value += rv.value/rv.count;
	      rvf.count += 1;
	    }
	}
    }
  
  

  bool ad = grid.adapt();

  Mapper.update();
  restrictionmap.reserve();
  data2.resize(Mapper.size());
  data.resize(Mapper.size());

  //---------------------------------//
  //MAPPING FROM PERSISTENT CONTAINER//
  //---------------------------------//
  for(int level = 0; level<=grid.maxLevel(); ++level)
    {
      LevelGridView levelView = grid.levelView(level);
      for(LevelElementIterator it = levelView.template begin<0>(); it != levelView.template end<0>(); ++it)
	{
	  if(! it->isNew())
	    {
	      if(it->isLeaf())
		{
		  RestrictedValue& rv = restrictionmap[*it];
		  int indexi = Mapper.map(*it);
		  data[indexi] = rv.value/rv.count;
		}
	    }
	  else
	    {
	      assert(it->level()>0);
	      ElementEntityPointer ep = it->father();
	      RestrictedValue& rvf = restrictionmap[*ep];
	      if(it->isLeaf())
		{
		  int indexi = Mapper.map(*it);
		  data[indexi] = rvf.value/rvf.count;
		}
	      else
		{
		  RestrictedValue& rv = restrictionmap[*it];
		  RestrictedValue& vr = restrictionmap2[*it];
		  rv.value = rvf.value/rvf.count;
		  rv.count = 1;
		}
	    }
	  
	  
	}      
    }
  
  grid.postAdapt();

  // std::vector<double> temp(data);
  // bool verify;

  //GHOST ELEMENTS UPDATE - NOT NECESSARY
  // VectorExchange<M,V> dh(Mapper,data,0);
  // grid.template communicate<VectorExchange<M,V> >(dh,Dune::InteriorBorder_All_Interface, Dune::ForwardCommunication);


  //  CHECK WHAT HAPPENS IN GHOST ELEMENTS
  // typedef typename LeafGridView::template Codim<0>::template Partition<Dune::Ghost_Partition>::Iterator GhostLeafIterator;
  // typedef typename GhostLeafIterator::Entity::Geometry GhostLeafGeometry;
  // GhostLeafIterator ghostendit = leafview.template end<0,Dune::Ghost_Partition>();
  // GhostLeafIterator ghostbeginit = leafview.template begin<0,Dune::Ghost_Partition>();
  // for(GhostLeafIterator ig = ghostbeginit; ig != ghostendit; ++ig){
  //   const GhostLeafGeometry& ggeo = ig->geometry();
  //   Dune::FieldVector<ct,dim> ggCenterCoords = ggeo.center();
  //   verify = data[Mapper.map(*ig)] == temp[Mapper.map(*ig)];
  //   std::cout << "GHOSTCHECK rank = " << helper.rank() << " coords = " << ggCenterCoords << "data(coords) = " << verify  << std::endl;
  //   //std::cout << "GHOSTCHECK rank = " << helper.rank() << " coords = " << ggCenterCoords << "data(coords) = "<< data[Mapper.map(*ig)] << std::endl;
  // }


  //PLOT data
  leafview = grid.leafView();
  Dune::VTKWriter<LeafGridView> vtkwriter(leafview);
  vtkwriter.addCellData(data,"data");
  std::string path="outputNEW";
  std::string extPath="";
  std::stringstream ss;
  ss << "UserData_" << plotCounter;
  vtkwriter.pwrite(ss.str(),path,extPath,Dune::VTKOptions::ascii);

  //PLOT data2
  ebegin = leafview.template begin<0>();
  eend = leafview.template end<0>();
  for(LeafElementIterator it = ebegin; it != eend; ++it)
    {
      int indexi = Mapper.map(*it);
      data2[indexi] = double(indexi);
    }
  Dune::VTKWriter<LeafGridView> vtkwriter2(leafview);
  vtkwriter2.addCellData(data2,"indices");
  std::stringstream dd;
  dd << "IndexData_" << plotCounter;
  vtkwriter2.pwrite(dd.str(),path,extPath,Dune::VTKOptions::ascii);


  return ad;
};

template <class T,int dim>
T modulo(Dune::FieldVector<T,dim> vettore)
{
  T mod = 0;
  for(int i = 0; i != dim; ++i)
    mod += vettore[i]*vettore[i];
  return std::sqrt(mod);
};

#endif
