#ifndef DEFORMATION2_HH
#define DEFORMATION2_HH

#include <dune/common/exceptions.hh>
#include <dune/grid/geometrygrid/coordfunction.hh>
#include <dune/grid/alugrid/3d/entity.hh>
//#include <dune/fem/space/common/functionspace.hh>
#include "myparametrization.hh"

#define kronecker(l,m) ((l)==(m)?1:0)


template <class G, class M, class V>
void computeParametrizations(G& grid, M& mapper, V& container);

template <class V>
V computePointOverBoundary(V& piudue, V& menouno);

template <class V>
V computeDerivatives(V& piu, V& meno);


// DeformationCoordFunction
// ------------------------

//! deformation depending on a discrete function 
template <class VertexVector, class G, class M, class V>
class DeformationDiscreteFunction2
  : public Dune::DiscreteCoordFunction< double, 3, DeformationDiscreteFunction2< VertexVector, G, M, V > > 
{

  typedef Dune::DiscreteCoordFunction< double, 3, DeformationDiscreteFunction2< VertexVector, G, M, V > > BaseType;
  typedef VertexVector RangeVector;

  friend void computeParametrizations<G, M, V>(G& grid, M& mapper, V& container);
  friend VertexVector computePointOverBoundary<VertexVector>(VertexVector& piudue, VertexVector& menouno);
  friend VertexVector computeDerivatives<VertexVector>(VertexVector& piu, VertexVector& meno);

public:

  //----------------//
  //  CONSTRUCTORS  //
  //----------------//
  
  //Empty constructor (to be deleted)
  DeformationDiscreteFunction2() {
    parametrizations_=0;
};

  //Constructor for parametrization provided by the user
  DeformationDiscreteFunction2(G& grid, V& parametrizations) : parametrizations_(parametrizations) {
    if(grid.comm().rank() == 0){
      std::cout << "Parametrizations have been provided by the user" << std::endl;
    }
    grid.comm().barrier();
  };
  
  //Main Constructor: it uses the base grid, the mapper and a container to compute and to store the parametrizations
  DeformationDiscreteFunction2(G& grid, M& mapper, V& parametrizations) : parametrizations_(parametrizations) {
    
    const int codim = 0;
    const Dune::PartitionIteratorType ptype = Dune::InteriorBorder_Partition;
    const int dim = G::dimension;
    typedef typename G::ctype ct;
    typedef typename G::LeafGridView LeafGridView;
    typedef typename LeafGridView::template Codim<codim>::template Partition<ptype>::Iterator LeafIterator;
    typedef typename LeafIterator::Entity LeafEntity;
    typedef typename G::template Codim<codim>::EntityPointer EntityPointer;
    typedef typename LeafEntity::Geometry LeafGeometry;
    typedef typename LeafGridView::IntersectionIterator LeafIntersectionIterator;
    
    RangeVector x,w;

    //-----------------------------------------------//
    // VERTICES AND DERIVATIVES STRUCTURE DEFINITION //
    //-----------------------------------------------//
    // Vertices and derivatives container. p has 4 indeces and contains 3D FieldVectors,
    // then the fifth index accesses the components of the point or derivative.
    // Index are u,v,w,d: u,v,w \in {-1,0,1,2} they define vertex in reference space,
    // while d \in {0,1,2,3} distinguish between vertices (0) and derivatives.
    // 0 means vertices, 1 derivatives along u, 2 along v and 3 along w.
    std::map<int,std::map<int, std::map<int,std::map<int, RangeVector> > > > pmap;

    // PERFORMANCE MAY BE OPTIMIZED USING VECTOR AND A POINTER, OFFSETTING THE INITIAL POSITION
    // USING NEGATIVE INDEX IS LEGAL WITH POINTERS, ACCESSING OUT-OF-BOUNDS IS NOT!!!


    //TESTVARIABLES
    ct testcomponent;
    int testIntersectionIndex;
    RangeVector testIntersectionCenter,testElementCenter;
    ct buffer;

    //-----------------------------------------------//
    // HERMITE THIRD ORDER INTERPOLANTS COEFFICIENTS //
    //-----------------------------------------------//
    // alpha is the structure collecting the coefficients (4 for every polynomials) of the 4 Hermite interpolants.
    // alpha[p][q][j] with p \in {0,1}, q \in {0,1} and j \in {1,2,3,4}:
    // first and second indices are for defining the interpolant
    // (p distinguishes between interpolants used for vertices (edges and faces, too) and the ones used for derivatives)
    // (q distinguishes between interpolants associated to the two extreme values of the parameters {0,1})
    // (j chooses one among 4 components of the )
    std::map<int,std::map<int,std::map<int,double> > > alpha;
    // first interpolant alpha^0_0
    alpha[0][0][0] = 1.0;
    alpha[0][0][1] = 0.0;
    alpha[0][0][2] = -3.0;
    alpha[0][0][3] = 2.0;
    // second interpolant alpha^0_1
    alpha[0][1][0] = 0.0;
    alpha[0][1][1] = 0.0;
    alpha[0][1][2] = 3.0;
    alpha[0][1][3] = -2.0;
    // third interpolant alpha^1_0
    alpha[1][0][0] = 0.0;
    alpha[1][0][1] = 1.0;
    alpha[1][0][2] = -2.0;
    alpha[1][0][3] = 1.0;
    // fourth interpolant alpha^1_1
    alpha[1][1][0] = 0.0;
    alpha[1][1][1] = 0.0;
    alpha[1][1][2] = -1.0;
    alpha[1][1][3] = 1.0;

    //---------------------------------------------//
    // COEFFICIENTS OF THE SECOND INTERPOLATION e2 //
    //---------------------------------------------//
    // LAMBDA contains the coefficients of the Q^3 polynomials obtained form the second 3D TFI
    // first 3D TFI is embedded in.
    std::map<int,std::map<int, std::map<int,RangeVector> > >  LAMBDA; 
    std::map<int,std::map<int, std::map<int,RangeVector> > >  GAMMA;
    for(int g=0; g<=3; ++g){
      for(int j=0; j<=3; ++j){
	for(int k=0; k<=3; ++k){
	  for(int i=0; i<=2; ++i){
	    LAMBDA[g][j][k][i] = 0.0;
	    GAMMA[g][j][k][i] = 0.0;
	  }
	}
      }
    }

    LeafGridView leafGridView = grid.leafView();

    // LOOP OVER ELEMENTS (codim 0 entities)
    LeafIterator endit = leafGridView.template end<codim,ptype>();
    for(LeafIterator it = leafGridView.template begin<codim,ptype>(); it != endit; ++it)
      {
	//INITIALIZE p-STRUCTURE to ZERO to USE THE *= OPERATOR
	for(int u=-1; u<=2; ++u){
	  for(int v=-1; v<=2; ++v){
	    for(int w=-1; w<=2; ++w){
	      for(int d=0; d<=3; ++d){
		for(int k=0; k<=2; ++k){
		  pmap[u][v][w][d][k] = 0;
		}
	      }
	    }
	  }
	}

	//ELEMENT VERTICES
	pmap[0][0][0][0]=it->geometry().corner(0);
	pmap[1][0][0][0]=it->geometry().corner(1);
	pmap[0][1][0][0]=it->geometry().corner(2);
	pmap[1][1][0][0]=it->geometry().corner(3);
	pmap[0][0][1][0]=it->geometry().corner(4);
	pmap[1][0][1][0]=it->geometry().corner(5);
	pmap[0][1][1][0]=it->geometry().corner(6);
	pmap[1][1][1][0]=it->geometry().corner(7);

	testElementCenter = it->geometry().center();
	
	//NEIGHBORS VERTICES
	LeafIntersectionIterator isend = leafGridView.iend(*it);
	for(LeafIntersectionIterator is = leafGridView.ibegin(*it); is != isend; ++is)
	  {
	    testIntersectionIndex = is->indexInInside();
	    testIntersectionCenter = is->geometry().center();
	    if((*is).neighbor())
	      {
		//testIntersectionIndex = is->indexInOutside();
		//------------------------------------//
		// AUXILIARY VERTICES FROM NEIGHBOURS //
		//------------------------------------//
		if(testIntersectionIndex == 0)
		  {
		    pmap[-1][0][0][0] = (*is).outside()->geometry().corner(0);
		    pmap[-1][1][0][0] = (*is).outside()->geometry().corner(2);
		    pmap[-1][0][1][0] = (*is).outside()->geometry().corner(4);
		    pmap[-1][1][1][0] = (*is).outside()->geometry().corner(6);
		  }
		if(testIntersectionIndex == 1)
		  {
		    pmap[2][0][0][0] = (*is).outside()->geometry().corner(1);
		    pmap[2][1][0][0] = (*is).outside()->geometry().corner(3);
		    pmap[2][0][1][0] = (*is).outside()->geometry().corner(5);
		    pmap[2][1][1][0] = (*is).outside()->geometry().corner(7);
		  }
		if(testIntersectionIndex == 2)
		  {
		    pmap[0][-1][0][0] = (*is).outside()->geometry().corner(0);
		    pmap[1][-1][0][0] = (*is).outside()->geometry().corner(1);
		    pmap[0][-1][1][0] = (*is).outside()->geometry().corner(4);
		    pmap[1][-1][1][0] = (*is).outside()->geometry().corner(5);
		  }
		if(testIntersectionIndex == 3)
		  {
		    pmap[0][2][0][0] = (*is).outside()->geometry().corner(2);
		    pmap[1][2][0][0] = (*is).outside()->geometry().corner(3);
		    pmap[0][2][1][0] = (*is).outside()->geometry().corner(6);
		    pmap[1][2][1][0] = (*is).outside()->geometry().corner(7);
		  }
		if(testIntersectionIndex == 4)
		  {
		    pmap[0][0][-1][0] = (*is).outside()->geometry().corner(0);
		    pmap[1][0][-1][0] = (*is).outside()->geometry().corner(1);
		    pmap[0][1][-1][0] = (*is).outside()->geometry().corner(2);
		    pmap[1][1][-1][0] = (*is).outside()->geometry().corner(3);
		  }
		if(testIntersectionIndex == 5)
		  {
		    pmap[0][0][2][0] = (*is).outside()->geometry().corner(4);
		    pmap[1][0][2][0] = (*is).outside()->geometry().corner(5);
		    pmap[0][1][2][0] = (*is).outside()->geometry().corner(6);
		    pmap[1][1][2][0] = (*is).outside()->geometry().corner(7);
		  }
	      }
	    else
	      {
		// NO NEIGHBOR AVAILABLE
		//------------------------------------------------------------------------------------//
		// AUXILIARY VERTICES COMPUTED TO PRESERVE THE DERIVATIVE OF THE OF THE LINEAR SEGMENT //
		//------------------------------------------------------------------------------------//
		if(testIntersectionIndex == 0)
		  {
		    pmap[-1][0][0][0] = computePointOverBoundary(pmap[0][0][0][0],pmap[1][0][0][0]);
		    pmap[-1][1][0][0] = computePointOverBoundary(pmap[0][1][0][0],pmap[1][1][0][0]);
		    pmap[-1][0][1][0] = computePointOverBoundary(pmap[0][0][1][0],pmap[1][0][1][0]);
		    pmap[-1][1][1][0] = computePointOverBoundary(pmap[0][1][1][0],pmap[1][1][1][0]);
		  }
		if(testIntersectionIndex == 1)
		  {
		    pmap[2][0][0][0] = computePointOverBoundary(pmap[1][0][0][0],pmap[0][0][0][0]);
		    pmap[2][1][0][0] = computePointOverBoundary(pmap[1][1][0][0],pmap[0][1][0][0]);
		    pmap[2][0][1][0] = computePointOverBoundary(pmap[1][0][1][0],pmap[0][0][1][0]);
		    pmap[2][1][1][0] = computePointOverBoundary(pmap[1][1][1][0],pmap[0][1][1][0]);
		  }
		if(testIntersectionIndex == 2)
		  {
		    pmap[0][-1][0][0] = computePointOverBoundary(pmap[0][0][0][0],pmap[0][1][0][0]);
		    pmap[1][-1][0][0] = computePointOverBoundary(pmap[1][0][0][0],pmap[1][1][0][0]);
		    pmap[0][-1][1][0] = computePointOverBoundary(pmap[0][0][1][0],pmap[0][1][1][0]);
		    pmap[1][-1][1][0] = computePointOverBoundary(pmap[1][0][1][0],pmap[1][1][1][0]);
		  }
		if(testIntersectionIndex == 3)
		  {
		    pmap[0][2][0][0] = computePointOverBoundary(pmap[0][1][0][0],pmap[0][0][0][0]);
		    pmap[1][2][0][0] = computePointOverBoundary(pmap[1][1][0][0],pmap[1][0][0][0]);
		    pmap[0][2][1][0] = computePointOverBoundary(pmap[0][1][1][0],pmap[0][0][1][0]);
		    pmap[1][2][1][0] = computePointOverBoundary(pmap[1][1][1][0],pmap[1][0][1][0]);
		  }
		if(testIntersectionIndex == 4)
		  {
		    pmap[0][0][-1][0] = computePointOverBoundary(pmap[0][0][0][0],pmap[0][0][1][0]);
		    pmap[1][0][-1][0] = computePointOverBoundary(pmap[1][0][0][0],pmap[1][0][1][0]);
		    pmap[0][1][-1][0] = computePointOverBoundary(pmap[0][1][0][0],pmap[0][1][1][0]);
		    pmap[1][1][-1][0] = computePointOverBoundary(pmap[1][1][0][0],pmap[1][1][1][0]);
		  }
		if(testIntersectionIndex == 5)
		  {
		    pmap[0][0][2][0] = computePointOverBoundary(pmap[0][0][1][0],pmap[0][0][0][0]);
		    pmap[1][0][2][0] = computePointOverBoundary(pmap[1][0][1][0],pmap[1][0][0][0]);
		    pmap[0][1][2][0] = computePointOverBoundary(pmap[0][1][1][0],pmap[0][1][0][0]);
		    pmap[1][1][2][0] = computePointOverBoundary(pmap[1][1][1][0],pmap[1][1][0][0]);
		  }
	      }
	    testIntersectionIndex = 0;
	  }
	
	//-------------//
	// DERIVATIVES //
	//-------------//
	// along u - pmap[u][v][w][1] with u,v,w \in {0,1}
	// along v - pmap[u][v][w][2] with u,v,w \in {0,1}
	// along w - pmap[u][v][w][3] with u,v,w \in {0,1}
	for(int u=0; u<=1; ++u){
	  for(int v=0; v<=1; ++v){
	    for(int w=0; w<=1; ++w){
	      pmap[u][v][w][1] = computeDerivatives(pmap[u+1][v][w][0],pmap[u-1][v][w][0]);
	      pmap[u][v][w][2] = computeDerivatives(pmap[u][v+1][w][0],pmap[u][v-1][w][0]);
	      pmap[u][v][w][3] = computeDerivatives(pmap[u][v][w+1][0],pmap[u][v][w-1][0]);
	    }
	  }
	}
	
	//-------------------------------------------------//
	// COMPUTE LAMBDA COEFFICIENTS - 2nd INTERPOLATION //
	//-------------------------------------------------//
	for(int g = 0; g <= 3; ++g){
	  for(int j = 0; j <= 3; ++j){
	    for(int k = 0; k <= 3; ++k){
	      for(int i = 0; i <= 2; ++i){
		
		for(int b = 0; b <= 1; ++b){
		  for(int l = 0; l <= 1; ++l){
		    for(int m = 0; m <= 1; ++m){

		      for(int p = 0; p <= 1; ++p){
			for(int q = 0; q <= 1; ++q){
			  if(p==0){
			    buffer = (pmap[b][q][l][0][i] * alpha[p][q][j] * alpha[0][l][k] * alpha[0][b][g] + pmap[b][m][q][0][i] * alpha[p][q][k] * alpha[0][m][j] * alpha[0][b][g]) * kronecker(l,m);
			  }
			  else if(p==1){
			    buffer = (pmap[b][q][l][2][i] * alpha[p][q][j] * alpha[0][l][k] * alpha[0][b][g] + pmap[b][m][q][3][i] * alpha[p][q][k] * alpha[0][m][j] * alpha[0][b][g]) * kronecker(l,m);
			  }
			  LAMBDA[g][j][k][i] += buffer;
			} // closing q-loop
		      } // closing p-loop
		      buffer = -pmap[b][l][m][0][i] * alpha[0][l][j] * alpha[0][m][k] * alpha[0][b][g];
		      LAMBDA[g][j][k][i] += buffer;

		      for(int p = 0; p <= 1; ++p){
			for(int q = 0; q <= 1; ++q){
			  if(p==0){
			    buffer = (pmap[q][b][l][0][i] * alpha[p][q][g] * alpha[0][l][k] * alpha[0][b][j] + pmap[m][b][q][0][i] * alpha[p][q][k] * alpha[0][m][g] * alpha[0][b][j]) * kronecker(l,m);
			  }
			  else if(p==1){
			    buffer = (pmap[q][b][l][1][i] * alpha[p][q][g] * alpha[0][l][k] * alpha[0][b][j] + pmap[m][b][q][3][i] * alpha[p][q][k] * alpha[0][m][g] * alpha[0][b][j]) * kronecker(l,m);
			  }
			  LAMBDA[g][j][k][i] += buffer;
			} // closing q-loop
		      } // closing p-loop
		      buffer = -pmap[l][b][m][0][i] * alpha[0][l][g] * alpha[0][m][k] * alpha[0][b][j];
		      LAMBDA[g][j][k][i] += buffer;

		    } // closing m-loop
		  } // closing l-loop
		} // closing b-loop

		for(int s = 0; s <= 3; ++s){
		  for(int b = 0; b <= 1; ++b){
		    for(int a = 0; a <= 1; ++a){
		      for(int l = 0; l <= 1; ++l){
			for(int m = 0; m <= 1; ++m){
			  
			  for(int p = 0; p <= 1; ++p){
			    for(int q = 0; q <= 1; ++q){
			      if(p==0){
				buffer = (-pmap[a][q][l][0][i] * alpha[p][q][s] * alpha[0][l][k] * alpha[0][a][g] * alpha[0][b][j] - pmap[a][m][q][0][i] * alpha[p][q][k] * alpha[0][m][s] * alpha[0][a][g] * alpha[0][b][j]) * kronecker(l,m) * pow(b,s);
			      }
			      else if(p==1){
				buffer = (-pmap[a][q][l][2][i] * alpha[p][q][s] * alpha[0][l][k] * alpha[0][a][g] * alpha[0][b][j] - pmap[a][m][q][3][i] * alpha[p][q][k] * alpha[0][m][s] * alpha[0][a][g] * alpha[0][b][j]) * kronecker(l,m) * pow(b,s);
			      }
			      LAMBDA[g][j][k][i] += buffer;
			    } // closing q-loop
			  } // closing p-loop
			  buffer = pmap[a][l][m][0][i] * alpha[0][l][s] * alpha[0][m][k] * alpha[0][a][g] * alpha[0][b][j] * pow(b,s);
			  LAMBDA[g][j][k][i] += buffer;

			} // closing m-loop
		      } // closing l-loop
		    } // closing a-loop
		  } // closing b-loop
		} // closing s-loop

	      } // closing i-loop
	    } // closing k-loop
	  } // closing j-loop
	} // closing g-loop
		      
	//------------------------------------------------------------------------------------//
	// COMPUTE GAMMA COEFFICIENTS - 3rd AND FINAL  INTERPOLATION //
	//------------------------------------------------------------------------------------//

	for(int g = 0; g <= 3; ++g){
	  for(int j = 0; j <= 3; ++j){
	    for(int k = 0; k <= 3; ++k){
	      for(int i = 0; i <= 2; ++i){
		
		GAMMA[g][j][k][i] = LAMBDA[g][j][k][i];
		
		for(int c = 0; c <= 1; ++c){
		  for(int l = 0; l <= 1; ++l){
		    for(int m = 0; m <= 1; ++m){

		      for(int p = 0; p <= 1; ++p){
			for(int q = 0; q <= 1; ++q){
			  if(p==0){
			    buffer = ( pmap[q][l][c][0][i] * alpha[p][q][g] * alpha[0][l][j] * alpha[0][c][k] + pmap[m][q][c][0][i] * alpha[p][q][j] * alpha[0][m][g] * alpha[0][c][k]) * kronecker(l,m);
			    GAMMA[g][j][k][i] += buffer;
			  }
			  else if(p==1){
			    buffer = ( pmap[q][l][c][1][i] * alpha[p][q][g] * alpha[0][l][j] * alpha[0][c][k] + pmap[m][q][c][2][i] * alpha[p][q][j] * alpha[0][m][g] * alpha[0][c][k]) * kronecker(l,m);
			    GAMMA[g][j][k][i] += buffer;
			  }
			}
		      }
		      buffer = -pmap[l][m][c][0][i] * alpha[0][l][g] * alpha[0][m][j] * alpha[0][c][k];
		      GAMMA[g][j][k][i] += buffer;
		    }
		  }
		}

		for(int h = 0; h <= 3; ++h){
		  for(int c = 0; c <= 1; ++c){
		    buffer = -alpha[0][c][k] * LAMBDA[g][j][h][i] * pow(c,h);
		    GAMMA[g][j][k][i] += buffer;
		  }
		}

	      }
	    }
	  }
	}
		    
	//-------------------------------------------------------------------------------//
	// BUILDING ELEMENT PARAMETRIZATION AND STORING IT IN PARAMETRIZATIONS ATTRIBUTE //
	//-------------------------------------------------------------------------------//
	MyParametrization<RangeVector,RangeVector> testparametrization(GAMMA);
	int indexi = mapper.map(*it);
	parametrizations[indexi] = testparametrization;



	//GARBAGE
	testcomponent=pmap[1][0][0][0][0];

	//EntityPointer father_ptr = it->father();
	//x=father_ptr->geometry().corner(0);
	
	x = it->geometry().corner(0);
	w = it->geometry().local(x);
	x = it->geometry().corner(1);
	w = it->geometry().local(x);
	x = it->geometry().corner(2);
	w = it->geometry().local(x);
	x = it->geometry().corner(3);
	w = it->geometry().local(x);
	x = it->geometry().corner(4);
	w = it->geometry().local(x);
	x = it->geometry().corner(5);
	w = it->geometry().local(x);
	x = it->geometry().corner(6);
	w = it->geometry().local(x);
	x = it->geometry().corner(7);
	w = it->geometry().local(x);
	//GARBAGE
      }

    //-----------------------------------------------//
    // COMMUNICATING PARAMETRIZATIONS TO GHOST CELLS //
    //-----------------------------------------------//
    VectorExchange<M,V> dhParam(mapper,parametrizations,0);
    grid.template communicate<VectorExchange<M,V> >(dhParam,Dune::InteriorBorder_All_Interface, Dune::ForwardCommunication);
  };

  //-------------------//
  //  EVALUATE METHOD  //
  //-------------------//

  template <class HostEntity> 
  void evaluate ( const HostEntity& entity, 
            		  unsigned int corner, 
                  RangeVector &y ) const
  {  
    // put the evaluation of the parametrization here
    RangeVector x = entity.geometry().corner(corner);
    RangeVector w = entity.geometry().local(x);

    //IDENTITY DEFORMATION
    y = entity.geometry().corner(corner);

    return;
  };


  //-------------------//
  //  ATTRIBUTES       //
  //-------------------//
  V parametrizations_;



};

template <class G, class M, class V>
void computeParametrizations(G& grid, M& mapper, V& container)
{
  double ciccio;
  return;
};

template <class V>
V computePointOverBoundary(V& piudue, V& menouno)
{
  V p;
  p = piudue;
  p *= 2;
  p -= menouno;
  return p;
};

template <class V>
V computeDerivatives(V& piu, V& meno)
{
  V p;
  p = piu;
  p -= meno;
  p *= 0.5;
  return p;
};


#endif
