#ifndef MY_BOUNDARY_SEGMENT_HH
#define MY_BOUNDARY_SEGMENT_HH

#include <dune/common/fvector.hh>
#include <dune/grid/common/boundarysegment.hh>
#include <dune/grid/common/boundaryprojection.hh>

/** \file
    \brief Base class for grid boundary segments of arbitrary geometry
*/

namespace Dune {

  /** \brief Base class for classes implementing geometries of boundary segments
      Some grid implementations, as for example UGGrid, allow to use boundary segments of
      arbitrary geometry.  That means that you can have grid boundaries approach smooth
      shapes when refining the grid.
      
      Such curved boundary segments are specified by giving classes that implement them.
      Each boundary segment is implemented by an object of a class derived from 
      BoundarySegment.  The set of these objects is handed over to the grid upon grid
      construction. 
  */
  template< int dim>
  class MyBoundarySegment : public Dune::BoundarySegment <dim>
  {
  public:
    /** \brief Dummy virtual destructor */
    //virtual ~MyBoundarySegment() {};
    
    MyBoundarySegment(const Dune::FieldVector<double,3>& center,double radius, double fromAngle1, double toAngle1, double fromAngle2, double toAngle2) : center_(center), radius_(radius), fromAngle1_(fromAngle1), toAngle1_(toAngle1), fromAngle2_(fromAngle2), toAngle2_(toAngle2) {}
    
    /** \brief A function mapping local coordinates on a boundary segment to world coordinates
     */
    Dune::FieldVector<double, dim> operator() ( const FieldVector< double, dim-1> &local ) const {
      Dune::FieldVector<double,dim> temp = center_;
      double angle1 = fromAngle1_ + local[0] * (toAngle1_ - fromAngle1_);// 0<=angle1<=2PI
      double angle2 = fromAngle2_ + local[1] * (toAngle2_ - fromAngle2_);// 0<=angle2<=PI
      temp[0] += radius_ * std::cos(angle1) * std::sin(angle2);
      temp[1] += radius_ * std::sin(angle1) * std::sin(angle2);
      temp[2] += radius_ * std::cos(angle2);
      return temp;

    }
    Dune::FieldVector<double,3> center_;
    double radius_;
    double fromAngle1_,fromAngle2_;
    double toAngle1_,toAngle2_;

  };


  template< int dim>
  class MyBoundaryProjection : public Dune::DuneBoundaryProjection <dim>
  {
  public:
    /** \brief Dummy virtual destructor */
    typedef FieldVector<double,dim> CoordinateType;
    MyBoundaryProjection() {}
    virtual ~MyBoundaryProjection() {};
    
    /** \brief A function mapping local coordinates on a boundary segment to world coordinates
     */
    virtual CoordinateType operator() ( const CoordinateType& global ) const {

      CoordinateType prj( global );
      //const double factor = radius_/global.two_norm();
      //prj*=factor;
      prj[2] = -4.0 + std::sqrt( (prj[0]-0.5)*(prj[0]-0.5) + (prj[1]-0.5)*(prj[1]-0.5));
      return prj;
    }

  };


  
  
  
}  // end namespace Dune
#endif
